//
//  MirrorView.m
//  MakeupMirror
//
//  Created by David Phillip Oster on 3/17/09.
//  Copyright 2009 David Phillip Oster. All rights reserved.
//

#import "MirrorView.h"
#import "CursorShielder.h"
#import "Fifo.h"
#import "ToolbarItem.h"

static NSString *const kMagnifyID = @"magnify";
static NSString *const kPauseID = @"pause";
static NSString *const kTimeDelayID = @"timeDelay";

enum {
  kFramesPerSecond = 30,
  kHistoryCapacity = 15*kFramesPerSecond
};

@interface MirrorView(Private)
- (void)connectToVideoGrabber;
- (void)disconnectFromVideoCapture;
- (void)startConnectAttemptTimer;
- (void)stopConnectAttemptTimer;
- (void)startTimeDelayPendingTimer;
- (void)stopTimeDelayPendingTimer;
- (void)update;
@end

@implementation MirrorView


- (id)initWithFrame:(NSRect)frame {
    self = [super initWithFrame:frame];
    if (self) {
      history_ = [[Fifo alloc] initWithCapacity:kHistoryCapacity];
      lock_ = [[NSLock alloc] init];
      toolBarIDs_ = [[NSArray alloc] initWithObjects:
        kMagnifyID,
        NSToolbarSpaceItemIdentifier,
        kPauseID,
        NSToolbarFlexibleSpaceItemIdentifier,
        kTimeDelayID,
        nil];
    }
    return self;
}

- (void)dealloc {
  NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
  [nc removeObserver:self];
  [shield_ invalidate];
  [shield_ release];
  shield_ = nil;
  [self disconnectFromVideoCapture];
  [self stopConnectAttemptTimer];
  [self stopTimeDelayPendingTimer];
  [pausedImage_ release];
  pausedImage_ = nil;
  CVBufferRelease(frameBuffer_);
  frameBuffer_ = nil;
  [lock_ release];
  lock_ = nil;
  [saveURL_ release];
  saveURL_ = nil;
  [toolBarIDs_ release];
  toolBarIDs_ = nil;
  [history_ release];
  history_ = nil;
  [super dealloc];
}

- (void)awakeFromNib {
  NSToolbar *toolbar = [[[NSToolbar alloc] initWithIdentifier:@"Toolbar"] autorelease];
  [toolbar setDelegate:self];
  [[self window] setToolbar:toolbar];

  if (nil == shield_) {
    shield_ = [[CursorShielder alloc] initWithView:self];
  }
  NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
  [nc addObserver:self selector:@selector(willClose:) name:NSWindowWillCloseNotification object:[self window]];
  [nc addObserver:self selector:@selector(willClose:) name:NSApplicationWillTerminateNotification object:NSApp];
  [nc addObserver:self selector:@selector(wasConnected:) name:QTCaptureDeviceWasConnectedNotification object:nil];
  [nc addObserver:self selector:@selector(wasDisconnected:) name:QTCaptureDeviceWasDisconnectedNotification object:nil];
  [self connectToVideoGrabber];
  [self startConnectAttemptTimer];

  NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
  NSNumber *mode = [ud objectForKey:@"mode"];
  if (mode) {
    NSToolbar *toolbar = [[self window] toolbar];
    [toolbar setDisplayMode:[mode intValue]];
    [toolbar setVisible:[ud boolForKey:@"toolbarIsVisible"]];
  }
  if ([ud boolForKey:kMagnifyID]) {
    [self toggleMagnify:self];
  }
  if ([ud boolForKey:kTimeDelayID]) {
    [self toggleTimeDelay:self];
  }
}

- (CIImage *)currentImageUnscaled {
  CVImageBufferRef buffer = frameBuffer_;
  CIImage *rawImage = nil;
  if (isPaused_ && pausedImage_) {
    rawImage = pausedImage_;
  } else {
    [lock_ lock];
    if (isTimeDelaying_ && nil == timeDelayPending_) {
      rawImage = [history_ removeObject];
    }
    if (nil == rawImage) {
      rawImage = [CIImage imageWithCVImageBuffer:buffer];
    }
    [lock_ unlock];
  }
  return rawImage;
}

- (CIImage*)currentImage {
  CIImage *rawImage = [self currentImageUnscaled];
  CGRect rawRect = [rawImage extent];
  NSRect bounds = [self bounds];
  float xScale = bounds.size.width/rawRect.size.width;
  float yScale = bounds.size.height/rawRect.size.height;
  float scaleFactor = fmax(xScale, yScale);
  if (isMagnifying_) {
    scaleFactor *= 2.;
  }
  CGAffineTransform transform = CGAffineTransformMakeScale(-scaleFactor, scaleFactor);
  return [rawImage imageByApplyingTransform:transform];
}

- (void)renderToCurrentContext:(CIImage *)frameImage {
  NSRect bounds = [self bounds];
  CGRect imageRect = [frameImage extent];
  NSGraphicsContext *currentContext = [NSGraphicsContext currentContext];
  CIContext *ciContext = [currentContext CIContext];
  if (isMagnifying_) {
    imageRect.size.width = (int) (imageRect.size.width / 2.);
    imageRect.size.height = (int) (imageRect.size.height/2.);
    imageRect.origin.x += (int) (imageRect.size.width/2.);
    imageRect.origin.y += (int) (imageRect.size.height/2.);
  }
  CGPoint topLeft = CGPointMake(
    (int)((bounds.size.width - imageRect.size.width) * 0.5),
    (int)((bounds.size.height - imageRect.size.height) * 0.5));
  [ciContext drawImage:frameImage atPoint:topLeft fromRect:imageRect];
}

- (void)drawMessage {
  NSString *mesg = NSLocalizedString(@"NoCamera",@"");
  NSString *info = [NSString stringWithFormat:@"\n%@",NSLocalizedString(@"NoCameraInfo",@"")];
  NSMutableParagraphStyle *style = [[[NSMutableParagraphStyle alloc] init] autorelease];
  [style setAlignment:NSCenterTextAlignment];
  NSFont *font = [NSFont boldSystemFontOfSize:24];
  NSShadow *shadow = [[[NSShadow alloc] init] autorelease];
  [shadow setShadowOffset:NSMakeSize(3, -3)];
  [shadow setShadowBlurRadius:6.];
  NSColor *color = [NSColor whiteColor];
  NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
    style, NSParagraphStyleAttributeName,
    color, NSForegroundColorAttributeName,
    font, NSFontAttributeName,
    shadow, NSShadowAttributeName,
    nil];
  NSMutableAttributedString *s = 
    [[[NSMutableAttributedString alloc] initWithString:mesg attributes:dict] autorelease];
  font = [NSFont boldSystemFontOfSize:17];
  [dict setObject:font forKey:NSFontAttributeName];
  NSMutableAttributedString *sInfo = 
    [[[NSMutableAttributedString alloc] initWithString:info attributes:dict] autorelease];
  NSRect bounds = [self bounds];
  bounds.size.height *= 0.66;
  [s appendAttributedString:sInfo];
  [s drawInRect:bounds];
}

- (void)drawRect:(NSRect)rect {
  NSRect bounds = [self bounds];
  if (captureSession_ && frameBuffer_) {
    CIImage *frameImage = [self currentImage];

    [[NSColor blackColor] set];
    [NSBezierPath fillRect:bounds];

    [self renderToCurrentContext:frameImage];
  } else {
    [[NSColor blueColor] set];
    [NSBezierPath fillRect:bounds];
    [self drawMessage];
  }
}


- (void)wasConnected:(NSNotification *)notify {
  if (nil != captureSession_ && nil != connectTimer_) {
    [self stopConnectAttemptTimer];
  }
}

- (void)startConnectAttemptTimer {
  if (nil == captureSession_) {
    connectTimer_ = [[NSTimer 
      scheduledTimerWithTimeInterval:4. 
                              target:self 
                            selector:@selector(tryConnectToVideoGrabberTimer:) 
                            userInfo:nil 
                             repeats:YES] retain];
  }
}

- (void)stopConnectAttemptTimer {
  [connectTimer_ invalidate];
  [connectTimer_ release];
  connectTimer_ = nil;
}

- (void)timeDelayNow:(NSTimer *)timer {
  [self stopTimeDelayPendingTimer];
  [self update];
}

- (void)startTimeDelayPendingTimer {
  [self stopTimeDelayPendingTimer];
  float secondsToDelay = (kHistoryCapacity - [history_ count]) /(float)kFramesPerSecond;
  if (secondsToDelay <= 0) {
    [self timeDelayNow:nil];
  } else {
    timeDelayPending_ = [[NSTimer 
        scheduledTimerWithTimeInterval:secondsToDelay 
                                target:self 
                              selector:@selector(timeDelayNow:) 
                              userInfo:nil 
                               repeats:NO] retain];
  }
}

- (void)stopTimeDelayPendingTimer {
  [timeDelayPending_ invalidate];
  [timeDelayPending_ release];
  timeDelayPending_ = nil;
}


- (void)disconnectFromVideoCapture {
  [captureSession_ release];
  captureSession_ = nil;
  [videoInput_ release];
  videoInput_ = nil;
  [videoOutput_ release];
  videoOutput_ = nil;
}

- (void)wasDisconnected:(NSNotification *)notify {
  [self disconnectFromVideoCapture];
  [self startConnectAttemptTimer];
  [self update];
}

- (void)update {
  [[self window] update];
  [self setNeedsDisplay:YES];
}

- (void)willClose:(NSNotification *)notify {
  if (!didClose_) {
    NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
    [ud setBool:isMagnifying_ forKey:kMagnifyID];
    [ud setBool:isTimeDelaying_ forKey:kTimeDelayID];

    NSToolbar *toolbar = [[self window] toolbar];
    [ud setBool:[toolbar isVisible] forKey:@"toolbarIsVisible"];
    int modeN = (int) [toolbar displayMode];
    [ud setObject:[NSNumber numberWithInt:modeN] forKey:@"mode"];

    [ud synchronize];

    NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
    [nc removeObserver:self name:NSWindowWillCloseNotification object:nil];
    [self disconnectFromVideoCapture];
    [self stopConnectAttemptTimer];
    [NSCursor setHiddenUntilMouseMoves:NO];
    didClose_ = YES;
    [shield_ invalidate];
  }
}

- (void)tryConnectToVideoGrabberTimer:(NSTimer *)timer {
  [self connectToVideoGrabber];
  if (nil != captureSession_) {
    [connectTimer_ invalidate];
    [connectTimer_ release];
    connectTimer_ = nil;
    [self update];
  }
}

- (void)captureOutput:(QTCaptureOutput *)captureOutput
  didOutputVideoFrame:(CVImageBufferRef)videoFrame
     withSampleBuffer:(QTSampleBuffer *)sampleBuffer 
       fromConnection:(QTCaptureConnection *)connection {

  [lock_ lock];
  if (frameBuffer_ != videoFrame) {
    CVBufferRelease(frameBuffer_);
    frameBuffer_ = videoFrame;
    CVBufferRetain(frameBuffer_);
  }
  CIImage *rawImage = [CIImage imageWithCVImageBuffer:videoFrame];
  if (rawImage) {
    [history_ addObject:rawImage];
  }
  [lock_ unlock];

  [self setNeedsDisplay:YES];
}


- (void)connectToVideoGrabber {
  captureSession_ = [[QTCaptureSession alloc] init];
  if (captureSession_) {
    NSError *error = nil;
    QTCaptureDevice *videoDevice = [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeVideo];
    if (nil == videoDevice) {
      videoDevice = [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeMuxed];
    }
    if( ! [videoDevice open:&error]) {
      videoDevice = nil;
    }
    if (videoDevice) {
      videoInput_ = [[QTCaptureDeviceInput alloc] initWithDevice:videoDevice];
    }
    if (videoInput_ && ![captureSession_ addInput:videoInput_ error:&error]) {
      [self disconnectFromVideoCapture];
    }
    if (captureSession_) {
      videoOutput_ = [[QTCaptureVideoPreviewOutput alloc] init];
      if (videoOutput_) {
        [videoOutput_ setDelegate:self];
        if(![captureSession_ addOutput:videoOutput_ error:&error]) {
          [self disconnectFromVideoCapture];
        }
      }
    }
    if (captureSession_ && videoInput_) {
      [self setCaptureSession:captureSession_];
      [captureSession_ startRunning];
    } else {
      [self disconnectFromVideoCapture];
    }
  }
  // Give enough time for a frame to come in, for magnify toolbar item to enable.
  [self performSelector:@selector(update) withObject:nil afterDelay:0.1];
}

- (void)saveImage:(CIImage *)currentImage url:(NSURL *)url {
  NSRect bounds = [self bounds];
  NSBitmapImageRep *bitmapImageRep = 
    [[[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL 
                                             pixelsWide:bounds.size.width 
                                             pixelsHigh:bounds.size.height 
                                          bitsPerSample:8 
                                        samplesPerPixel:4 
                                               hasAlpha:YES 
                                               isPlanar:NO 
                                         colorSpaceName:NSDeviceRGBColorSpace 
                                            bytesPerRow:0 
                                           bitsPerPixel:0] autorelease];
  NSGraphicsContext *nsContext = [NSGraphicsContext graphicsContextWithBitmapImageRep:bitmapImageRep];
  [NSGraphicsContext saveGraphicsState];
  [NSGraphicsContext setCurrentContext:nsContext];
  [self renderToCurrentContext:currentImage];
  [NSGraphicsContext restoreGraphicsState];
  NSDictionary *info = [NSDictionary dictionaryWithObjectsAndKeys: 
    [NSNumber numberWithFloat:0.7], NSImageCompressionFactor,
    nil];
  NSData *outputData = [bitmapImageRep representationUsingType:NSJPEGFileType properties:info];
  [outputData writeToURL:url atomically:YES];
}

#pragma mark -

- (BOOL)magnifyOn {
  return isMagnifying_;
}

- (void)setMagnifyOn:(BOOL)magnifyOn {
  if (magnifyOn != isMagnifying_) {
    [self toggleMagnify:self];
  }
}


- (BOOL)timeDelayOn {
  return isTimeDelaying_ && nil == timeDelayPending_;
}

- (void)setTimeDelayOn:(BOOL)timeDelayOn {
  if (timeDelayOn != isTimeDelaying_) {
    [self toggleTimeDelay:self];
  }
}


- (BOOL)pauseOn {
  return isPaused_;
}

- (void)setPauseOn:(BOOL)pauseOn {
  if (pauseOn != isPaused_) {
    [self togglePause:self];
  }
}

- (BOOL)videoOn {
  return nil != captureSession_;
}


#pragma mark -

- (IBAction)saveDocument:(id)sender {
  if (saveURL_) {
    [self saveImage:[self currentImage] url:saveURL_];
  } else {
    [self saveDocumentAs:sender];
  }
}

- (void)savePanelDidEnd:(NSSavePanel *)sheet returnCode:(NSInteger)returnCode contextInfo:(void *)contextInfo {
  CIImage *currentImage = (CIImage *)contextInfo;
  if (returnCode == NSOKButton) {
    NSURL *url = [sheet URL];
    if (url != saveURL_) {
      [saveURL_ release];
      saveURL_ = [url retain];
    }
    [self saveImage:currentImage url:url];
  }
  [currentImage release];
}

- (IBAction)saveDocumentAs:(id)sender {
  NSSavePanel *savePanel = [NSSavePanel savePanel];
  [savePanel setDelegate:self];
  [savePanel setAllowedFileTypes:[NSArray arrayWithObjects:@"jpg", nil]];

  NSString *path = nil;
  NSString *file = nil;
  if (saveURL_) {
    path = [[saveURL_ path] stringByDeletingLastPathComponent];
    file = [[saveURL_ path] lastPathComponent];
  }
  [savePanel setCanSelectHiddenExtension:YES];
  [savePanel beginSheetForDirectory:path file:file modalForWindow:[self window] modalDelegate:self didEndSelector:@selector(savePanelDidEnd:returnCode:contextInfo:) contextInfo:[[self currentImage] retain]];
}


- (IBAction)toggleMagnify:(id)sender {
  isMagnifying_ = ! isMagnifying_;
  [self update];
}

- (IBAction)togglePause:(id)sender {
  if (!isPaused_) {
    pausedImage_ = [[self currentImageUnscaled] retain];
  } else {
    [lock_ lock];
    [pausedImage_ release];
    pausedImage_ = nil;
    [lock_ unlock];
  }
  isPaused_ = ! isPaused_;
  [self update];
}

- (IBAction)toggleTimeDelay:(id)sender {
  isTimeDelaying_ = ! isTimeDelaying_;
  if(isPaused_) {
    [self togglePause:sender];
  }
  [self startTimeDelayPendingTimer];
}

- (BOOL)validateUserInterfaceItem:(NSObject<NSValidatedUserInterfaceItem> *)menuItem {
  SEL sel = [menuItem action];
  if (sel == @selector(toggleMagnify:)) {
    if ([menuItem respondsToSelector:@selector(setState:)]) {
      [(NSMenuItem *)menuItem setState:isMagnifying_ ? NSOnState : NSOffState];
    }
    return nil != frameBuffer_;
  } else if (sel == @selector(togglePause:)) {
    if ([menuItem respondsToSelector:@selector(setState:)]) {
      [(NSMenuItem *)menuItem setState:isPaused_ ? NSOnState : NSOffState];
    }
    return nil != captureSession_;
  } else if (sel == @selector(toggleTimeDelay:)) {
    if ([menuItem respondsToSelector:@selector(setState:)]) {
      int state = NSOffState;
      if (isTimeDelaying_) {
        state = timeDelayPending_ ? NSMixedState : NSOnState;
      }
      [(NSMenuItem *)menuItem setState:state];
    }
    return nil != captureSession_;
  } else if (sel == @selector(saveDocument:) ||
      sel == @selector(saveDocumentAs:)) {

    return nil != frameBuffer_;
  } else if (sel == @selector(undo:) ||
      sel == @selector(redo:)) {

    return NO;
  }
  return YES;
}

- (BOOL)validateToolbarItem:(NSToolbarItem *)toolBarItem {
  return [self validateUserInterfaceItem:toolBarItem];
}

- (NSToolbarItem *)toolbar:(NSToolbar *)toolbar
     itemForItemIdentifier:(NSString *)itemIdentifier
 willBeInsertedIntoToolbar:(BOOL)flag {
  NSToolbarItem *item = [[[ToolbarItem alloc] initWithItemIdentifier:itemIdentifier] autorelease];

  if ([itemIdentifier isEqual:kMagnifyID]) {
    [item setLabel:NSLocalizedString(kMagnifyID, @"")];
    [item setImage:[NSImage imageNamed:kMagnifyID]];
    [item setTag:1];
    [item setTarget:self];
    [item setAction:@selector(toggleMagnify:)];
  } else if ([itemIdentifier isEqual:kPauseID]) {
    [item setLabel:NSLocalizedString(kPauseID, @"")];
    [item setImage:[NSImage imageNamed:kPauseID]];
    [item setTarget:self];
    [item setAction:@selector(togglePause:)];
    [item setTag:2];
  } else if ([itemIdentifier isEqual:kTimeDelayID]) {
    [item setLabel:NSLocalizedString(kTimeDelayID, @"")];
    [item setImage:[NSImage imageNamed:kTimeDelayID]];
    [item setTarget:self];
    [item setAction:@selector(toggleTimeDelay:)];
    [item setTag:3];
  }
  return item;
}

- (NSArray *)toolbarDefaultItemIdentifiers:(NSToolbar*)toolbar {
  return toolBarIDs_;
}

- (NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar*)toolbar {
  return toolBarIDs_;
}

- (NSArray *)toolbarSelectableItemIdentifiers:(NSToolbar *)toolbar {
  return nil;
}


@end
